from charset_normalizer.constant import UNICODE_RANGES_ZIP
from functools import lru_cache


class UnicodeRangeIdentify:

    @staticmethod
    @lru_cache(maxsize=8192)
    def find_letter_type(letter):
        """
        This method is intended to associate a single character with a range name from the unicode table
        :param str letter: Shall be a unique char
        :return: Associated unicode range designation
        :rtype: Union[str, None]
        """
        if len(letter) != 1:
            raise IOError('Trying to associate multiple char <{}> to a single unicode range'.format(letter))

        for u_name, u_range in UNICODE_RANGES_ZIP.items():

            if ord(letter) in u_range:
                return u_name

        return None

    @staticmethod
    @lru_cache(maxsize=8192)
    def is_accentuated(letter):
        """
        Verify if a latin letter is accentuated, unicode point of view.
        :param str letter: Letter to check
        :return: True if accentuated, else False
        :rtype: bool
        """
        if len(letter) != 1:
            raise IOError('Trying to determine accentuated state of multiple char <{}>'.format(letter))
        return 192 <= ord(letter) <= 383



import re
import statistics
from encodings.aliases import aliases
from os.path import basename, splitext
import collections

from cached_property import cached_property

from charset_normalizer.probe_coherence import ProbeCoherence
from charset_normalizer.probe_chaos import ProbeChaos

from platform import python_version_tuple


class CharsetNormalizerMatch:

    RE_PRINTABLE_LETTER = re.compile(r'[0-9\W\n\r\t]+')

    def __init__(self, b_content, guessed_source_encoding, chaos_ratio, ranges):
        """
        :param bytes b_content: Raw binary content
        :param str guessed_source_encoding: Guessed source encoding accessible by Python
        :param float chaos_ratio: Coefficient of previously detected mess in decoded content
        """

        self._raw = b_content
        self._encoding = guessed_source_encoding
        self._chaos_ratio = chaos_ratio

        self._string = str(self._raw, encoding=self._encoding).replace('\r', '')

        self.char_counter = collections.Counter(re.sub(CharsetNormalizerMatch.RE_PRINTABLE_LETTER, '', self._string.lower()))

        self.ranges = ranges

    @cached_property
    def w_counter(self):
        return collections.Counter(re.sub(CharsetNormalizerMatch.RE_PRINTABLE_LETTER, ' ', self._string.lower()).split())

    @cached_property
    def alphabets(self):
        """
        Discover list of alphabet in decoded content
        :return: List of alphabet
        :rtype: list[str]
        """
        return list(self.ranges.keys())

    @cached_property
    def could_be_from_charset(self):
        """
        Return list of possible originating charset
        :return: list of encoding
        :rtype: list[str]
        """
        return [self.encoding]

    def __eq__(self, other):
        """
        :param CharsetNormalizerMatch other:
        :return:
        """
        return self.chaos == other.chaos and len(self.raw) == len(other.raw) and self.encoding == other.encoding

    @cached_property
    def coherence(self):
        """
        Return a value between 0. and 1.
        Closest to 0. means that the initial string is considered coherent,
        Closest to 1. means that the initial string SEEMS NOT coherent.
        :return: Ratio as floating number
        :rtype: float
        """
        return ProbeCoherence(self.char_counter).ratio

    @cached_property
    def chaos(self):
        """
        Return a value between 0. and 1.
        Closest to 1. means that the initial string is considered as chaotic,
        Closest to 0. means that the initial string SEEMS NOT chaotic.
        :return: Ratio as floating number
        :rtype: float
        """
        return self._chaos_ratio

    @cached_property
    def chaos_secondary_pass(self):
        """
        Check once again chaos in decoded text, except this time, with full content.
        :return:
        """
        return ProbeChaos(str(self))

    @property
    def encoding(self):
        """
        :return: Guessed possible/probable originating charset
        :rtype: str
        """
        return self._encoding

    @property
    def raw(self):
        return self._raw

    def first(self):
        """
        Just in case
        :return: himself
        :rtype: CharsetNormalizerMatch
        """
        return self

    def best(self):
        """
        Just in case
        :return: himself
        :rtype: CharsetNormalizerMatch
        """
        return self

    def __str__(self):
        return self._string

    def output(self, encoding='utf-8'):
        """
        :param encoding:
        :return:
        :rtype: bytes
        """
        return str(self).encode(encoding)


class CharsetNormalizerMatches:

    def __init__(self, matches):
        """
        :param list[CharsetNormalizerMatch] matches:
        """
        self._matches = matches

    def __iter__(self):
        for elem in self._matches:
            yield elem

    def __getitem__(self, item):
        if not isinstance(item, int):
            raise KeyError
        return self._matches[item]

    def __len__(self):
        return len(self._matches)

    @staticmethod
    def normalize(path, steps=10, chunk_size=512, threshold=0.09):
        """
        :param str path:
        :param int steps:
        :param int chunk_size:
        :param float threshold:
        :return:
        """

        matches = CharsetNormalizerMatches.from_path(
            path,
            steps,
            chunk_size,
            threshold
        )

        b_name = basename(path)
        b_name_ext = list(splitext(b_name))

        if len(matches) == 0:
            raise IOError('Unable to normalize "{}", no encoding charset seems to fit.'.format(b_name))

        b_ = matches.best().first()

        b_name_ext[0] += '-' + b_.encoding

        with open('{}'.format(path.replace(b_name, ''.join(b_name_ext))), 'w', encoding='utf_8') as fp:
            fp.write(str(b_))

        return b_

    @staticmethod
    def from_bytes(sequences, steps=10, chunk_size=512, threshold=0.09):
        """
        Take a sequence of bytes that could potentially be decoded to str and discard all obvious non supported
        charset encoding.
        :param bytearray sequences: Actual sequence of bytes to analyse
        :param float threshold: Maximum amount of chaos allowed on first pass
        :param int chunk_size: Size to extract and analyse in each step
        :param int steps: Number of steps
        :return: List of potential matches
        :rtype: CharsetNormalizerMatches
        """
        py_v = [int(el) for el in python_version_tuple()]
        py_need_sort = py_v[0] < 3 or (py_v[0] == 3 and py_v[1] < 6)

        supported = sorted(aliases.items()) if py_need_sort else aliases.items()
        tested = set()
        working = dict()

        maximum_length = len(sequences)

        for support in supported:

            k, p = support

            if p in tested:
                continue

            tested.add(p)

            try:
                str(sequences, encoding=p)
            except UnicodeDecodeError:
                continue
            except LookupError:
                continue

            chaos_measures = list()
            ranges_encountered_t = dict()
            decoded_len_t = 0

            for i in range(0, maximum_length, int(maximum_length / steps)):

                chunk = sequences[i:i + chunk_size]
                decoded = str(chunk, encoding=p, errors='ignore')

                probe_chaos = ProbeChaos(decoded)
                chaos_measure, ranges_encountered = probe_chaos.ratio, probe_chaos.encountered_unicode_range_occurrences

                for k, e in ranges_encountered.items():
                    if k not in ranges_encountered_t.keys():
                        ranges_encountered_t[k] = 0
                    ranges_encountered_t[k] += e

                if chaos_measure > threshold:
                    if p in working.keys():
                        del working[p]
                    break

                chaos_measures.append(chaos_measure)

                if p not in working.keys():
                    working[p] = dict()

            if p in working.keys():
                working[p]['ratio'] = statistics.mean(chaos_measures)
                working[p]['ranges'] = ranges_encountered_t
                working[p]['chaos'] = sum(chaos_measures)
                working[p]['len'] = decoded_len_t

            if p == 'ascii' and working[p]['ratio'] == 0.:
                break

        return CharsetNormalizerMatches(
            [CharsetNormalizerMatch(sequences, enc, working[enc]['ratio'], working[enc]['ranges']) for enc in
             (sorted(working.keys()) if py_need_sort else working.keys()) if working[enc]['ratio'] <= threshold])

    @staticmethod
    def from_fp(fp, steps=10, chunk_size=512, threshold=0.09):
        """
        :param io.BinaryIO fp:
        :param int steps:
        :param int chunk_size:
        :param float threshold:
        :return:
        """
        return CharsetNormalizerMatches.from_bytes(
            bytearray(fp.read()),
            steps,
            chunk_size,
            threshold
        )

    @staticmethod
    def from_path(path, steps=10, chunk_size=512, threshold=0.09):
        """
        :param str path:
        :param int steps:
        :param int chunk_size:
        :param float threshold:
        :return:
        """
        with open(path, 'rb') as fp:
            return CharsetNormalizerMatches.from_fp(fp, steps, chunk_size, threshold)

    @cached_property
    def could_be_from_charset(self):
        """
        Return list of possible originating charset
        :return: list of encoding
        :rtype: list[str]
        """
        return [el.encoding for el in self._matches]

    def first(self):
        """
        Select first match available
        :return: first match available
        :rtype: CharsetNormalizerMatch or None
        """
        return self._matches[0] if len(self._matches) > 0 else None

    def best(self):
        """
        Keep only the matches with the lowest ratio of chaos and check eventually for coherence to keep the best of the
        best matches.
        :return: Single match or list of matches
        :rtype: CharsetNormalizerMatches | CharsetNormalizerMatch
        """

        lowest_ratio = None
        lowest_ratio_frequency = None

        match_per_ratio = dict()
        match_per_frequency_letter = dict()

        for match in self._matches:

            if match.chaos not in match_per_ratio.keys():
                match_per_ratio[match.chaos] = list()

            match_per_ratio[match.chaos].append(match)

            if lowest_ratio is None or lowest_ratio > match.chaos:
                lowest_ratio = match.chaos

        if lowest_ratio is None:
            return CharsetNormalizerMatches([])

        all_latin_basic = True

        for match in match_per_ratio[lowest_ratio]:  # type: CharsetNormalizerMatch
            secondary_ratio = match.coherence

            if lowest_ratio_frequency is None or lowest_ratio_frequency > secondary_ratio:
                lowest_ratio_frequency = secondary_ratio

            if secondary_ratio not in match_per_frequency_letter.keys():
                match_per_frequency_letter[secondary_ratio] = list()

            match_per_frequency_letter[secondary_ratio].append(match)

            if len(match.alphabets) != 1 or match.alphabets[0] != 'Latin de base':
                all_latin_basic = False

        if all_latin_basic is True:
            return CharsetNormalizerMatches(match_per_frequency_letter[lowest_ratio_frequency]).first()

        return CharsetNormalizerMatches(match_per_frequency_letter[lowest_ratio_frequency]) if len(match_per_frequency_letter[lowest_ratio_frequency]) > 1 else CharsetNormalizerMatches(match_per_frequency_letter[lowest_ratio_frequency]).first()



"""
This file contain, for each range a name associated with for Unicode.
Scrapped from https://unicode-table.com/
"""
UNICODE_RANGES = [
    "0000−001F",
    "0020−007F",
    "0080−00FF",
    "0100−017F",
    "0180−024F",
    "0250−02AF",
    "02B0−02FF",
    "0300−036F",
    "0370−03FF",
    "0400−04FF",
    "0500−052F",
    "0530−058F",
    "0590−05FF",
    "0600−06FF",
    "0700−074F",
    "0750−077F",
    "0780−07BF",
    "07C0−07FF",
    "0800−083F",
    "0840−085F",
    "0860−086F",
    "08A0−08FF",
    "0900−097F",
    "0980−09FF",
    "0A00−0A7F",
    "0A80−0AFF",
    "0B00−0B7F",
    "0B80−0BFF",
    "0C00−0C7F",
    "0C80−0CFF",
    "0D00−0D7F",
    "0D80−0DFF",
    "0E00−0E7F",
    "0E80−0EFF",
    "0F00−0FFF",
    "1000−109F",
    "10A0−10FF",
    "1100−11FF",
    "1200−137F",
    "1380−139F",
    "13A0−13FF",
    "1400−167F",
    "1680−169F",
    "16A0−16FF",
    "1700−171F",
    "1720−173F",
    "1740−175F",
    "1760−177F",
    "1780−17FF",
    "1800−18AF",
    "18B0−18FF",
    "1900−194F",
    "1950−197F",
    "1980−19DF",
    "19E0−19FF",
    "1A00−1A1F",
    "1A20−1AAF",
    "1AB0−1AFF",
    "1B00−1B7F",
    "1B80−1BBF",
    "1BC0−1BFF",
    "1C00−1C4F",
    "1C50−1C7F",
    "1C80−1C8F",
    "1CC0−1CCF",
    "1CD0−1CFF",
    "1D00−1D7F",
    "1D80−1DBF",
    "1DC0−1DFF",
    "1E00−1EFF",
    "1F00−1FFF",
    "2000−206F",
    "2070−209F",
    "20A0−20CF",
    "20D0−20FF",
    "2100−214F",
    "2150−218F",
    "2190−21FF",
    "2200−22FF",
    "2300−23FF",
    "2400−243F",
    "2440−245F",
    "2460−24FF",
    "2500−257F",
    "2580−259F",
    "25A0−25FF",
    "2600−26FF",
    "2700−27BF",
    "27C0−27EF",
    "27F0−27FF",
    "2800−28FF",
    "2900−297F",
    "2980−29FF",
    "2A00−2AFF",
    "2B00−2BFF",
    "2C00−2C5F",
    "2C60−2C7F",
    "2C80−2CFF",
    "2D00−2D2F",
    "2D30−2D7F",
    "2D80−2DDF",
    "2DE0−2DFF",
    "2E00−2E7F",
    "2E80−2EFF",
    "2F00−2FDF",
    "2FF0−2FFF",
    "3000−303F",
    "3040−309F",
    "30A0−30FF",
    "3100−312F",
    "3130−318F",
    "3190−319F",
    "31A0−31BF",
    "31C0−31EF",
    "31F0−31FF",
    "3200−32FF",
    "3300−33FF",
    "3400−4DBF",
    "4DC0−4DFF",
    "4E00−9FFF",
    "A000−A48F",
    "A490−A4CF",
    "A4D0−A4FF",
    "A500−A63F",
    "A640−A69F",
    "A6A0−A6FF",
    "A700−A71F",
    "A720−A7FF",
    "A800−A82F",
    "A830−A83F",
    "A840−A87F",
    "A880−A8DF",
    "A8E0−A8FF",
    "A900−A92F",
    "A930−A95F",
    "A960−A97F",
    "A980−A9DF",
    "A9E0−A9FF",
    "AA00−AA5F",
    "AA60−AA7F",
    "AA80−AADF",
    "AAE0−AAFF",
    "AB00−AB2F",
    "AB30−AB6F",
    "AB70−ABBF",
    "ABC0−ABFF",
    "AC00−D7AF",
    "D7B0−D7FF",
    "D800−DB7F",
    "DB80−DBFF",
    "DC00−DFFF",
    "E000−F8FF",
    "F900−FAFF",
    "FB00−FB4F",
    "FB50−FDFF",
    "FE00−FE0F",
    "FE10−FE1F",
    "FE20−FE2F",
    "FE30−FE4F",
    "FE50−FE6F",
    "FE70−FEFF",
    "FF00−FFEF",
    "FFF0−FFFF",
    "10000−1007F",
    "10080−100FF",
    "10100−1013F",
    "10140−1018F",
    "10190−101CF",
    "101D0−101FF",
    "10280−1029F",
    "102A0−102DF",
    "102E0−102FF",
    "10300−1032F",
    "10330−1034F",
    "10350−1037F",
    "10380−1039F",
    "103A0−103DF",
    "10400−1044F",
    "10450−1047F",
    "10480−104AF",
    "104B0−104FF",
    "10500−1052F",
    "10530−1056F",
    "10600−1077F",
    "10800−1083F",
    "10840−1085F",
    "10860−1087F",
    "10880−108AF",
    "108E0−108FF",
    "10900−1091F",
    "10920−1093F",
    "10980−1099F",
    "109A0−109FF",
    "10A00−10A5F",
    "10A60−10A7F",
    "10A80−10A9F",
    "10AC0−10AFF",
    "10B00−10B3F",
    "10B40−10B5F",
    "10B60−10B7F",
    "10B80−10BAF",
    "10C00−10C4F",
    "10C80−10CFF",
    "10E60−10E7F",
    "11000−1107F",
    "11080−110CF",
    "110D0−110FF",
    "11100−1114F",
    "11150−1117F",
    "11180−111DF",
    "111E0−111FF",
    "11200−1124F",
    "11280−112AF",
    "112B0−112FF",
    "11300−1137F",
    "11400−1147F",
    "11480−114DF",
    "11580−115FF",
    "11600−1165F",
    "11660−1167F",
    "11680−116CF",
    "11700−1173F",
    "118A0−118FF",
    "11A00−11A4F",
    "11A50−11AAF",
    "11AC0−11AFF",
    "11C00−11C6F",
    "11C70−11CBF",
    "11D00−11D5F",
    "12000−123FF",
    "12400−1247F",
    "12480−1254F",
    "13000−1342F",
    "14400−1467F",
    "16800−16A3F",
    "16A40−16A6F",
    "16AD0−16AFF",
    "16B00−16B8F",
    "16F00−16F9F",
    "16FE0−16FFF",
    "17000−187FF",
    "18800−18AFF",
    "1B000−1B0FF",
    "1B100−1B12F",
    "1B170−1B2FF",
    "1BC00−1BC9F",
    "1BCA0−1BCAF",
    "1D000−1D0FF",
    "1D100−1D1FF",
    "1D200−1D24F",
    "1D300−1D35F",
    "1D360−1D37F",
    "1D400−1D7FF",
    "1D800−1DAAF",
    "1E000−1E02F",
    "1E800−1E8DF",
    "1E900−1E95F",
    "1EE00−1EEFF",
    "1F000−1F02F",
    "1F030−1F09F",
    "1F0A0−1F0FF",
    "1F100−1F1FF",
    "1F200−1F2FF",
    "1F300−1F5FF",
    "1F600−1F64F",
    "1F650−1F67F",
    "1F680−1F6FF",
    "1F700−1F77F",
    "1F780−1F7FF",
    "1F800−1F8FF",
    "1F900−1F9FF",
    "20000−2A6DF",
    "2A700−2B73F",
    "2B740−2B81F",
    "2B820−2CEAF",
    "2CEB0−2EBEF",
    "2F800−2FA1F",
    "E0000−E007F",
    "E0100−E01EF"
]

UNICODE_RANGES_NAMES = [
    "Commandes C0",
    "Latin de base",
    "Supplément latin-1",
    "Latin étendu A",
    "Latin étendu B",
    "Alphabet phonétique international (IPA)",
    "Lettres modificatives avec chasse",
    "Diacritiques",
    "Grec et copte",
    "Cyrillique",
    "Suppl. cyrillique",
    "Arménien",
    "Hébreu",
    "Arabe",
    "Syriaque",
    "Supplément arabe",
    "Thâna",
    "N’ko",
    "Samaritain",
    "Mandéen",
    "Syriac Supplement",
    "Arabe étendu A",
    "Devanâgarî",
    "Bengalî",
    "Gourmoukhî",
    "Goudjarâtî",
    "Oriyâ",
    "Tamoul",
    "Télougou",
    "Kannara",
    "Malayalam",
    "Singhalais",
    "Thaï",
    "Lao",
    "Tibétain",
    "Birman",
    "Géorgien",
    "Jamos hangûl",
    "Éthiopien",
    "Suppl. éthiopien",
    "Tchérokî",
    "Syllabaires autochtones canadiens",
    "Ogam",
    "Runes",
    "Tagal",
    "Hanounóo",
    "Bouhid",
    "Tagbanoua",
    "Khmer",
    "Mongol",
    "Syllab. autochtones canadiens étendus",
    "Limbu",
    "Taï-le",
    "Nouveau taï lü",
    "Symb. khmers",
    "Bougui",
    "Taï tham",
    "Combining Diacritical Marks Extended",
    "Balinais",
    "Sundanais",
    "Batak",
    "Lepcha",
    "Ol chiki",
    "Cyrillic Extended C",
    "Sund.",
    "Vedic Extensions",
    "Supplément phonétique",
    "Suppl. phonétique étendu",
    "Supplément de diacritiques",
    "Latin étendu additionnel",
    "Grec étendu",
    "Ponctuation générale",
    "Exposants et indices",
    "Symboles monétaires",
    "Signes comb. p. symb.",
    "Symboles de type lettre",
    "Formes numérales",
    "Flèches",
    "Symboles mathématiques",
    "Signes techniques divers",
    "Pictogrammes de commande",
    "Recon. optique",
    "Alphanumériques cerclés",
    "Filets",
    "Pavés",
    "Formes géométriques",
    "Symboles divers",
    "Casseau",
    "Div. symb. math.-A",
    "S. A fl.",
    "Combinaisons Braille",
    "Supplément B de flèches",
    "Divers symboles mathématiques - B",
    "Opérateurs mathématiques supplémentaires",
    "Divers symboles et flèches",
    "Glagolitique",
    "Latin étendu C",
    "Copte",
    "Supplément géorgien",
    "Tifinaghe",
    "Éthiopien étendu",
    "Cyrillique ét. A",
    "Ponctuation complémentaire",
    "Formes supplémentaires des clés CJC",
    "Clés chinoises kangxies",
    "Dsc. id.",
    "Symboles et ponct. CJC",
    "Hiragana",
    "Katakana",
    "Bopomofo",
    "Jamos de compatibilité hangûl",
    "Kanbn.",
    "Bopomofo ét.",
    "Traits CJC",
    "S. kana",
    "Lettres et mois CJC cerclés",
    "Compatibilité CJC",
    "Supplément A aux idéogrammes unifiés CJC",
    "Hexagrammes du Yi Jing",
    "Idéogrammes unifiés CJC",
    "Syllabaire yi des Monts frais",
    "Clés yi",
    "Lisu",
    "Vaï",
    "Cyrillique étendu B",
    "Bamoum",
    "L. modif. de ton",
    "Latin étendu D",
    "Sylotî nâgrî",
    "N° Ind.",
    "Phags-pa",
    "Saurashtra",
    "Dévanâgarî ét.",
    "Kayah li",
    "Rejang",
    "Jamos hangûl-A",
    "Javanais",
    "Myanmar Extended-B",
    "Cham",
    "Birman ét.-A",
    "Taï việt",
    "Meitei. ét.",
    "Éthiopien étendu A",
    "Latin Extended-E",
    "Cherokee Supplement",
    "Meitei mayek",
    "Hangûl",
    "Hangul Jamo Extended-B",
    "Demi-zone haute d’indirection à usage privé",
    "High Private Use Surrogates",
    "Demi-zone basse d’indirection",
    "Zone à usage privé",
    "Idéogrammes de compatibilité CJC",
    "Formes de prés. alphabétiques",
    "F. A p.a.",
    "Sél.var.",
    "Vertic.",
    "½-dia.",
    "F. comp. CJC",
    "Petites formes",
    "Formes B de présentation arabes",
    "Formes de demi et pleine chasse",
    "Sp.",
    "Syllabaire linéaire B",
    "Idéogrammes du linéaire B",
    "Nombres égéens",
    "Nombres grecs anciens",
    "Symboles anciens",
    "Disque de Phaïstos",
    "Lycien",
    "Carien",
    "N° epact coptes",
    "Alphabet italique",
    "Gotique",
    "(ancien permien)",
    "Ougaritique",
    "Vieux perse",
    "Déséret",
    "Shavien",
    "Osmanya",
    "Osage",
    "Elbasan",
    "Aghbanie",
    "Linéaire A",
    "Syllabaire chypriote",
    "Araméen impérial",
    "Palmyrénien",
    "Nabatéen",
    "Hatran",
    "Phénicien",
    "Lydien",
    "Hiéroglyphes méroïtiques",
    "Cursif méroïtique",
    "Kharochthî",
    "Sud-arabique",
    "Nord-arabique",
    "Manichéen",
    "Avestique",
    "Parthe des inscriptions",
    "Pehlevi des inscriptions",
    "(pehelevi des psautiers)",
    "Alphabet de l’Orkhon",
    "Old Hungarian",
    "N° rûmi",
    "Brahmî",
    "Kaïthî",
    "Sora sompeng",
    "Chakma",
    "(mahajani)",
    "Sharada",
    "N°cingalais",
    "Khojkî",
    "Multani",
    "Khudawadi",
    "(grantha)",
    "Newa",
    "Tirhouta",
    "(siddham)",
    "(modi)",
    "Mongolian Supplement",
    "Takrî",
    "Ahom",
    "Warang Citi",
    "Zanabazar Square",
    "Soyombo",
    "Pau cin hau",
    "Bhaiksuki",
    "Marchen",
    "Masaram Gondi",
    "Cunéiforme",
    "Ponctuation et nombres cunéiformes",
    "Early Dynastic Cuneiform",
    "Hiéroglyphes égyptiens",
    "Anatolian Hieroglyphs",
    "Supplément bamoum",
    "Mro",
    "Alphabet bassa",
    "Pahawh hmong",
    "Ahmao",
    "Ideographic Symbols and Punctuation",
    "Tangut",
    "Tangut Components",
    "Supplément kana",
    "Kana Extended-A",
    "Nushu",
    "Duployan",
    "Shorthand Format Controls",
    "Symboles musicaux byzantins",
    "Symboles musicaux occidentaux",
    "Notation musicale grecque ancienne",
    "Symboles du Classique du mystère suprême",
    "Chiffres-bâtonnets chinois",
    "Symboles mathématiques alphanumériques",
    "Sutton SignWriting",
    "Glagolitic Supplement",
    "Mende Kikakui",
    "Adlam",
    "Arabic Mathematical Alphabetic Symbols",
    "Pièces de mah-jong",
    "Dominos",
    "Cartes à jouer",
    "Supplément alphanumérique cerclé",
    "Supplément idéographique cerclé",
    "Divers symboles et pictogrammes",
    "Émoticônes",
    "Casseau ornem.",
    "Symboles du transport et cartographiques",
    "Symboles alchimiques",
    "Formes géométriques étendues",
    "Supplément de flèches - C",
    "Supplemental Symbols and Pictographs",
    "CJK Unified Ideographs Extension B",
    "CJK Unified Ideographs Extension C",
    "CJK Unified Ideographs Extension D",
    "CJK Unified Ideographs Extension E",
    "CJK Unified Ideographs Extension F",
    "CJK Compatibility Ideographs Supplement",
    "Tags",
    "Variation Selectors Supplement"
]

UNICODE_RANGES_ZIP = dict(
    zip(
        UNICODE_RANGES_NAMES,
        [
            range(
                int(el.split('−')[0], 16),
                int(el.split('−')[1], 16)+1
            ) for el in UNICODE_RANGES
        ]
    )
)




from charset_normalizer.normalizer import CharsetNormalizerMatches, CharsetNormalizerMatch
from charset_normalizer.unicode import UnicodeRangeIdentify
from charset_normalizer.probe_chaos import ProbeChaos
from charset_normalizer.probe_coherence import ProbeCoherence



import re

from charset_normalizer.unicode import UnicodeRangeIdentify

from dragonmapper.hanzi import identify as s_identify
from dragonmapper.hanzi import MIXED, BOTH, UNKNOWN
from zhon.hanzi import sentence as cjc_sentence_re

from functools import lru_cache


@lru_cache(maxsize=8192)
class ProbeChaos:

    def __init__(self, string):
        """
        :param str string:
        """

        if not isinstance(string, str):
            raise TypeError('Cannot probe chaos from type <{}>, expected <str>'.format(type(string)))

        self._string = string

        self.successive_upper_lower = 0
        self.successive_accent = 0
        self.successive_different_unicode_range = 0
        self.encountered_unicode_range = set()
        self.unprintable = 0
        self.encountered_white_space = 0
        self.not_encountered_white_space = 0

        self.encountered_unicode_range_occurrences = dict()

        self.not_encountered_white_space_reset = 0

        self.previous_printable_letter = None

        self.previous_encountered_unicode_range = None

        self.total_letter_encountered = 0

        self.total_lower_letter_encountered = 0
        self.total_upper_accent_encountered = 0
        self.total_upper_accent_encountered_inner = 0
        self.total_unaccented_letter_encountered = 0

        self.gave_up = False

        if len(self._string) >= 10:
            self._probe()

    def __dict__(self):
        return {
            'successive_upper_lower': self.successive_upper_lower,
            'successive_accent': self.successive_accent,
            'successive_different_unicode_range': self.successive_different_unicode_range,
            'encountered_unicode_range': self.encountered_unicode_range,
            'unprintable': self.unprintable,
            'encountered_white_space': self.encountered_white_space,
            'not_encountered_white_space': self.not_encountered_white_space,
            'encountered_unicode_range_occurrences': self.encountered_unicode_range_occurrences,
            'not_encountered_white_space_reset': self.not_encountered_white_space_reset,
            'total_letter_encountered': self.total_letter_encountered,
            '_unravel_suspicious_ranges': self._unravel_suspicious_ranges(),
            'total_lower_letter_encountered': self.total_lower_letter_encountered,
            'total_upper_accent_encountered': self.total_upper_accent_encountered,
            'total_upper_accent_encountered_inner': self.total_upper_accent_encountered_inner,
            'total_unaccented_letter_encountered': self.total_unaccented_letter_encountered,
            'cjc_id': s_identify(self._string) if 'Idéogrammes unifiés CJC' in self.encountered_unicode_range else 0,
            'f(x)': '({} + {} + {} + {} + {} + {}) / {}'.format(self.successive_upper_lower, self.successive_accent, self.successive_different_unicode_range, self.not_encountered_white_space, self.unprintable, self._unravel_suspicious_ranges(), len(self._string))
        }

    def _unravel_suspicious_ranges(self):
        len_ = len(self._string)
        s_ = 0

        for k, v in self.encountered_unicode_range_occurrences.items():
            k_ = k.lower()
            if ('latin' not in k_ and 'ponctuation générale' not in k_ and 'formes de demi et pleine chasse' not in k_ and 'symboles et ponct.' not in k_ and 'cjc' not in k_) or 'latin étendu' in k_ or 'supplément latin' in k_:
                if v / len_ < 0.09:
                    if len(self.encountered_unicode_range_occurrences.keys()) <= 2 and 'supplément latin' in k_:
                        continue
                    # print('Suspicous ranges added', k, v)
                    s_ += v if 'formes géométriques' not in k_ else v*10

        return s_

    def _probe(self):

        c__ = False

        # Todo: Gain 200ms by doing this : list(zip(self._string, self._string[1:]))[::2]
        # OR zip(self._string[::2], self._string[1::2])
        for c, i_ in zip(self._string, range(0, len(self._string))):

            if not c__ and (i_ / len(self._string)) > 0.2 and self.ratio >= 0.3:
                self.gave_up = True
                break
            elif c__ is False and (i_ / len(self._string)) > 0.2:
                c__ = True

            self.total_letter_encountered += 1

            if not c.isprintable():
                if c not in ['\n', '\t', '\r']:
                    u_name = UnicodeRangeIdentify.find_letter_type(c)
                    if 'CJC' not in u_name and 'Ponctuation générale' not in u_name and ord(c) != 160:  # CJC have there own space char
                        self.unprintable += 2
                else:
                    self.encountered_white_space += 1
                    self.not_encountered_white_space = 0
                    self.not_encountered_white_space_reset += 1
                continue

            if c.isspace():
                self.encountered_white_space += 1
                self.not_encountered_white_space = 0
                self.not_encountered_white_space_reset += 1
                self.previous_printable_letter = c
                continue

            if self.not_encountered_white_space_reset < 2:
                self.not_encountered_white_space += 1

            if self.previous_printable_letter is None:
                self.previous_printable_letter = c
                continue

            is_accent = UnicodeRangeIdentify.is_accentuated(c)
            u_name = UnicodeRangeIdentify.find_letter_type(c)
            u_name_lower = u_name.lower() if u_name is not None else None

            if u_name is not None and u_name not in self.encountered_unicode_range:
                self.encountered_unicode_range_occurrences[u_name] = 0
                self.encountered_unicode_range.add(u_name)

            if is_accent and UnicodeRangeIdentify.is_accentuated(self.previous_printable_letter):
                self.successive_accent += 2

            if c.islower():
                self.total_lower_letter_encountered += 1

            if c.isupper() and is_accent:
                self.total_upper_accent_encountered += 1

                if self.previous_printable_letter.isalpha():
                    self.total_upper_accent_encountered_inner += 1
            elif not is_accent and c.isalpha():
                self.total_unaccented_letter_encountered += 1

            if u_name is not None:
                self.encountered_unicode_range_occurrences[u_name] += 1

                if 'symboles et ponct.' in u_name_lower or 'ponctuation générale' in u_name_lower:
                    self.encountered_white_space += 1
                    self.not_encountered_white_space = 0
                    self.not_encountered_white_space_reset += 1

                if 'latin' in u_name_lower or 'formes de demi et pleine chasse' in u_name_lower or 'symboles et ponct.' in u_name_lower or 'ponctuation générale' in u_name_lower or u_name is None:
                    self.previous_printable_letter = c
                    continue
                elif (self.previous_printable_letter.isupper() and c.islower()) or (
                        self.previous_printable_letter.islower() and c.isupper()):
                    self.successive_upper_lower += 1

                if u_name != self.previous_encountered_unicode_range and self.previous_encountered_unicode_range is not None:
                    if 'latin' not in self.previous_encountered_unicode_range.lower() and \
                            'formes de demi et pleine chasse' not in self.previous_encountered_unicode_range.lower() and \
                            'symboles et ponct.' not in self.previous_encountered_unicode_range.lower() and \
                            'ponctuation générale' not in self.previous_encountered_unicode_range.lower():
                        # print('successive unicode range', u_name, 'with', previous_encountered_unicode_range)
                        self.successive_different_unicode_range += 1

            self.previous_encountered_unicode_range = u_name
            self.previous_printable_letter = c

    def _unravel_cjc_suspicious(self):
        if 'Idéogrammes unifiés CJC' in self.encountered_unicode_range:
            if s_identify(self._string) in [MIXED, BOTH]:
                return self.encountered_unicode_range_occurrences['Idéogrammes unifiés CJC']
            elif len(re.findall(cjc_sentence_re, self._string)) == 0:
                return self.encountered_unicode_range_occurrences['Idéogrammes unifiés CJC']

        return UNKNOWN

    @property
    def ratio(self):
        """
        Return a value between 0. and 1.
        Closest to 1. means that the initial string is considered as chaotic,
        Closest to 0. means that the initial string SEEMS NOT chaotic.
        :return: Ratio as floating number
        :rtype: float
        """
        r_ = self.total_upper_accent_encountered if self.total_letter_encountered > 0 and self.total_unaccented_letter_encountered / self.total_letter_encountered < 0.5 else 0
        return (r_ + self.successive_upper_lower + self.successive_accent + self.successive_different_unicode_range + self.not_encountered_white_space + self.unprintable + self._unravel_suspicious_ranges() + self._unravel_cjc_suspicious()) / len(self._string)  # + len(self.encountered_unicode_range)-1




import unittest
from charset_normalizer.probe_chaos import ProbeChaos


class TestProbeChaos(unittest.TestCase):

    def test_not_gibberish(self):

        self.assertEqual(
            ProbeChaos('典肇乎庚辰年十二月廿一，及己丑年二月十九，收各方語言二百五十，合逾七百萬目；二十大卷佔八成，單英文卷亦過二百萬。悉文乃天下有志共筆而成；有意助之，幾網路、隨纂作，大典茁焉。').ratio,
            0.
        )

        self.assertEqual(
            ProbeChaos('العقلية , التنويم المغناطيسي و / أو الاقتراح').ratio,
            0.
        )

        self.assertEqual(
            ProbeChaos("RadoZ تـــعــــديــل الـــتــــوقــيــــت مـــن قــبــل").ratio,
            0.
        )

    def test_subtle_gibberish(self):

        self.assertLessEqual(
            ProbeChaos("Cehennemin Sava■þ²s²'da kim?").ratio,
            0.5
        )

        self.assertGreaterEqual(
            ProbeChaos("Cehennemin Sava■þ²s²'da kim?").ratio,
            0.
        )

        self.assertGreater(
            ProbeChaos('´Á¥½³ø§i --  ±i®Ìºû, ³¯·Ø©v').ratio,
            0.
        )

        self.assertLessEqual(
            ProbeChaos("´Á¥½³ø§i --  ±i®Ìºû, ³¯·Ø©v").ratio,
            0.5
        )

        self.assertGreater(
            ProbeChaos("ïstanbul, T■rkiye'nin en kalabal»k, iktisadi ve k■lt■rel aÓ»dan en —nemli").ratio,
            0.
        )

        self.assertLessEqual(
            ProbeChaos("ïstanbul, T■rkiye'nin en kalabal»k, iktisadi ve k■lt■rel aÓ»dan en —nemli").ratio,
            0.5
        )

        self.assertLessEqual(
            ProbeChaos("<i>Parce que Óa, c'est la vÕritable histoire de la rencontre avec votre Tante Robin.</i>").ratio,
            0.5
        )

    def test_complete_gibberish(self):
        self.assertTrue(
            ProbeChaos("""ØĢØŠØģØ§ØĶŲ ŲŲ ØĢŲ Ø§ŲŲØ§Øģ ŲŲŲ ŲØ§ ØģŲŲŲØŠØģØ§ØĶŲŲŲØ ØŊØđŲØ§ ŲØģŲØđ ØđŲ (ŲØąŲØŊŲ) ŲØ§ŲØŪØ§ØŠŲ""").gave_up,
        )

        self.assertTrue(
            ProbeChaos("""ÇáÚŞáíÉ , ÇáÊäæíã ÇáãÛäÇØíÓí æ / Ãæ ÇáÇŞÊÑÇÍ""").gave_up,
        )

    def test_part_gibberish(self):

        self.assertGreater(
            ProbeChaos(
                """hishamkoc@yahoo.com ุชุฑุฌูููุฉ ููุดูููุงู ุงููููููููุงูRadoZ ุชูููุนููููุฏูููู ุงููููุชูููููููููููููุช ููููู ูููุจููู""").ratio,
            0.4
        )

        self.assertGreater(
            ProbeChaos("锌褉械锌芯写邪胁邪褌械谢褟屑懈 锌芯褝褌芯 ").ratio,
            0.4
        )


if __name__ == '__main__':
    unittest.main()



import unittest
from charset_normalizer.unicode import UnicodeRangeIdentify


class TestUnicodeHelper(unittest.TestCase):
    def test_should_be_accented(self):

        self.assertTrue(
            UnicodeRangeIdentify.is_accentuated('é')
        )

        self.assertTrue(
            UnicodeRangeIdentify.is_accentuated('è')
        )

        self.assertTrue(
            UnicodeRangeIdentify.is_accentuated('è')
        )

        self.assertTrue(
            UnicodeRangeIdentify.is_accentuated('à')
        )

        self.assertTrue(
            UnicodeRangeIdentify.is_accentuated('À')
        )

        self.assertTrue(
            UnicodeRangeIdentify.is_accentuated('Ù')
        )

        self.assertTrue(
            UnicodeRangeIdentify.is_accentuated('ç')
        )

    def test_should_throw(self):
        with self.assertRaises(IOError):
            UnicodeRangeIdentify.is_accentuated('àé')
        with self.assertRaises(IOError):
            UnicodeRangeIdentify.is_accentuated('aé')
        with self.assertRaises(IOError):
            UnicodeRangeIdentify.is_accentuated('aa')


if __name__ == '__main__':
    unittest.main()



import unittest
from charset_normalizer.normalizer import CharsetNormalizerMatches
from charset_normalizer.probe_coherence import ProbeCoherence
from collections import Counter


class TestProbeCoherence(unittest.TestCase):
    def test_obvious_coherence_gap(self):

        should_be_most_coherent = CharsetNormalizerMatches.from_path('./data/sample.1.ar.srt').best().first().coherence

        with open('./data/sample.1.ar.srt', 'r', encoding='mac_cyrillic') as fp:
            r_ = ProbeCoherence(Counter(fp.read())).ratio

        with open('./data/sample.1.ar.srt', 'r', encoding='cp1251') as fp:
            t_ = ProbeCoherence(Counter(fp.read())).ratio

        self.assertLess(
            should_be_most_coherent,
            r_
        )

        self.assertLess(
            should_be_most_coherent,
            t_
        )


if __name__ == '__main__':
    unittest.main()



import statistics
from collections import Counter
from hashlib import sha256

from cached_property import cached_property

import json
from os.path import dirname, realpath, exists


class ProbeCoherence:

    FREQUENCIES = None
    CACHED = dict()
    ASSETS_PATH = '{}/assets'.format(dirname(realpath(__file__)))

    def __init__(self, character_occurrences):
        """
        :param collections.Counter character_occurrences:
        """
        if not isinstance(character_occurrences, Counter):
            raise TypeError('Cannot probe coherence from type <{}>, expected <collections.Counter>'.format(type(character_occurrences)))

        if ProbeCoherence.FREQUENCIES is None:
            with open('{}/frequencies.json'.format(ProbeCoherence.ASSETS_PATH) if exists('{}/frequencies.json'.format(ProbeCoherence.ASSETS_PATH)) else './charset_normalizer/assets/frequencies.json', 'r') as fp:
                ProbeCoherence.FREQUENCIES = json.load(fp)

        self._character_occurrences = character_occurrences
        self.most_common = character_occurrences.most_common()
        self.nb_count_occurrence = sum(character_occurrences.values())
        self.rank_per_lang = dict()

        if self.uuid not in ProbeCoherence.CACHED:
            self._probe()

    @cached_property
    def uuid(self):
        return sha256(''.join(list(self._character_occurrences.keys())).encode('utf-8')).hexdigest()

    @property
    def ratio(self):
        """
        Return a value between 0. and 1.
        Closest to 0. means that the initial string is considered coherent,
        Closest to 1. means that the initial string SEEMS NOT coherent.
        :return: Ratio as floating number
        :rtype: float
        """
        if self.uuid not in ProbeCoherence.CACHED:
            ProbeCoherence.CACHED[self.uuid] = statistics.mean([c for l, c in self.rank_per_lang.items()]) if len(self.rank_per_lang.keys()) > 0 else 1.
        return ProbeCoherence.CACHED[self.uuid]

    def _probe(self):
        for language, letters in ProbeCoherence.FREQUENCIES.items():

            most_common_cpy = list()
            n_letter_not_available = 0

            for o_letter, o_appearances in self.most_common:

                if o_letter.lower() not in letters and o_appearances/self.nb_count_occurrence >= 0.0009:
                    n_letter_not_available += 1
                elif o_appearances/self.nb_count_occurrence >= 0.0009:
                    most_common_cpy.append(
                        (o_letter.lower(), o_appearances)
                    )

            if len(most_common_cpy) == 0:
                continue

            ratio_unavailable_letters = n_letter_not_available/len(self.most_common)

            if ratio_unavailable_letters < 0.4:
                not_respected_rank_coeff, n_tested_on = self._verify_order_on(letters, most_common_cpy)

                if not_respected_rank_coeff < 0.7:
                    self.rank_per_lang[language] = not_respected_rank_coeff

    @staticmethod
    def _verify_order_on(target_alphabet_ordered, character_occurrences):
        """
        Verify if a particular ordered set of character correspond more or less to our list of character occurrences
        :param list[str] target_alphabet_ordered:
        :param list[tuple[str, int]] character_occurrences:
        :return: Compute a ratio, Closest to 1. means our occurrences does not respect at all our ordered alphabet and closest to 0. means the opposite
        """
        n_not_rightfully_ranked = 0
        n_tested = 0

        n_letter_alphabet = len(target_alphabet_ordered)

        letters_failed = list()

        for w_index, (w_l, w_r) in zip(range(0, len(character_occurrences)), character_occurrences):
            if w_l not in target_alphabet_ordered:
                continue

            n_tested += 1

            r_index = target_alphabet_ordered.index(w_l)

            r_min_s, r_max_s = w_index - int(0.15 * n_letter_alphabet), w_index + int(0.15 * n_letter_alphabet)

            if r_min_s < 0:
                r_min_s = 0
            if r_max_s > n_letter_alphabet - 1:
                r_max_s = n_letter_alphabet - 1

            if not r_min_s <= r_index <= r_max_s:
                letters_failed.append((w_l, w_r))
                n_not_rightfully_ranked += 1
            elif r_index == w_index and n_not_rightfully_ranked > 0:
                n_not_rightfully_ranked -= 1

        if n_tested == 0:
            return 1., 0

        return (n_not_rightfully_ranked / n_tested) if n_tested >= 22 else 1., n_tested
